@@ -2,11 +2,16 @@
 # -*- coding: utf-8 -*-
 """Fun-Fact"""
 from .backend import use, active_backend, available_backends
-from .lang import index, indices, tensor, template, _0, _1, delta
+from .lang import (
+    index, indices, tensor, template, pi, zeros, ones, eye
+)
 from .lang._math import *  # noqa: F401, F403
 from .model import Factorization
 from .algorithm import factorize
-from .context import set_optimizable
+from .vectorization import vectorize, view
+from .context import is_grad_on, enable_grad
+from . import initializers
+from . import conditions
 
 
 __all__ = [
@@ -17,16 +22,22 @@ __all__ = [
     'indices',
     'tensor',
     'template',
-    '_0',
-    '_1',
-    'delta',
+    'zeros',
+    'ones',
+    'eye',
+    'pi',
     'Factorization',
     'factorize',
-    'set_optimizable',
+    'vectorize',
+    'view',
+    'is_grad_on',
+    'enable_grad',
+    'initializers',
+    'conditions'
 ]
 
 
-__version__ = '0.7.1'
+__version__ = '0.9.1'
 __author__ = '''Yu-Hang "Maxin" Tang, Daan Camps, Elizaveta Rebrova'''
 __maintainer__ = 'Yu-Hang "Maxin" Tang'
 __email__ = 'Tang.Maxin@gmail.com'
