@@ -1,9 +1,10 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 import warnings
-from ._tsrex import index, indices, tensor
+from ._tsrex import index, indices, tensor, TsrEx
 from ._tplex import template
-from ._predefined_literal import _0, _1, delta
+from ._constants import pi
+from ._special import zeros, ones, eye
 
 
 try:
@@ -22,4 +23,14 @@ except Exception:
     warnings.warn('Cannot set up MathJAX, LaTeX rendering may not work.')
 
 
-__all__ = ['index', 'indices', 'tensor', 'template', '_0', '_1', 'delta']
+__all__ = [
+    'index',
+    'indices',
+    'tensor',
+    'template',
+    'zeros',
+    'ones',
+    'eye',
+    'pi',
+    'TsrEx'
+]
