@@ -1,19 +1,24 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
-from copy import copy
 import dataclasses
 import re
 import sys
 import asciitree
 import functools
+from numbers import Number
+from plum import Dispatcher
 from funfact.backend import active_backend as ab
 from funfact.util.iterable import as_namedtuple, as_tuple, flatten_if
 from ._ast import _AST, _ASNode, Primitives as P
 from .interpreter import (
-    dfs_filter, ASCIIRenderer, LatexRenderer, IndexPropagator, ShapeAnalyzer,
-    EinsteinSpecGenerator
+    ASCIIRenderer,
+    Compiler,
+    EinsteinSpecGenerator,
+    IndexnessAnalyzer,
+    LatexRenderer,
 )
-from ._terminal import AbstractIndex, AbstractTensor
+from ._terminal import LiteralValue, AbstractIndex, AbstractTensor
+from funfact.conditions import NoCondition
 
 
 class ASCIITreeFactory:
@@ -38,8 +43,13 @@ class ASCIITreeFactory:
                         )
                     )
                 ),
-                get_text=lambda node: node.ascii + ' ' + ' '.join([
-                    f'({v}: {_getattr_safe(node, v)})' for v in extra_fields
+                get_text=lambda node: '{}: {} '.format(
+                    node.name, node.ascii
+                ) + ' '.join([
+                    '({f}: {s})'.format(
+                        f=f,
+                        s=re.sub('\n', '', str(_getattr_safe(node, f)))
+                    ) for f in extra_fields
                 ])
             ),
             draw=asciitree.drawing.BoxStyle(
@@ -85,175 +95,255 @@ class _BaseEx(_AST):
 
     _latex_intr = LatexRenderer()
     _asciitree_factory = ASCIITreeFactory()
+    _compiler = Compiler()
     _einspec_generator = EinsteinSpecGenerator()
-    _index_propagator = IndexPropagator()
-    _shape_analyzer = ShapeAnalyzer()
+    _indexness_analyzer = IndexnessAnalyzer()
 
     @functools.lru_cache()
     def _repr_html_(self):
+        '''LaTeX rendering of the tensor expression for Jupyter notebook.'''
         return f'''$${self._latex_intr(self.root)}$$'''
 
     @property
     def asciitree(self):
+        '''A ASCII-based tree visualization of the tensor expression.
+        This property can also be treated as a callable that accepts a list of
+        string-valued attribute names to be displayed for each node.
+
+        Examples:
+            >>> import funfact as ff
+            >>> a = ff.tensor('a', 2, 3)
+            >>> b = ff.tensor('b', 3, 5)
+            >>> i, j, k = ff.indices('i, j, k')
+            >>> tsrex = a[i, j] * b[j, k]
+            >>> tsrex.asciitree
+            sum:multiply
+            ├── a[i,j]
+            │   ├──
+            │   ╰── i,j
+            │       ├── i
+            │       ╰── j
+            ╰── b[j,k]
+                ├── b
+                ╰── j,k
+                    ├── j
+                    ╰── k
+
+            >>> fac = ff.Factorization.from_tsrex(tsrex)
+            >>> fac.tsrex.asciitree('data')
+            sum:multiply (data: None)
+            ├── a[i,j] (data: None)
+            │   ├── a (data: [[ 0.573 -0.406  0.164]
+            │   │             [-0.492  0.128  1.428]])
+            │   ╰── i,j (data: None)
+            │       ├── i (data: None)
+            │       ╰── j (data: None)
+            ╰── b[j,k] (data: None)
+                ├── b (data: [[-0.527  0.933  0.218 -1.49  -0.976]
+                │             [ 0.028  0.285 -0.701 -1.753 -2.313]
+                │             [-0.371 -0.406  0.374  1.17   1.372]])
+                ╰── j,k (data: None)
+                    ├── j (data: None)
+                    ╰── k (data: None)
+        '''
         return self._asciitree_factory(self.root)
 
     @property
     @functools.lru_cache()
     def _static_analyzed(self):
-        return self._einspec_generator(
-            self._shape_analyzer(self._index_propagator(self.root))
-        )
+        return (self |
+                self._indexness_analyzer |
+                self._compiler |
+                self._einspec_generator).root
 
     @property
     def shape(self):
+        '''Shape of the evaluation result'''
         return self._static_analyzed.shape
 
-    @property
-    def live_indices(self):
-        return self._static_analyzed.live_indices
-
     @property
     def ndim(self):
+        '''Dimensionality of the evaluation result'''
         return len(self._static_analyzed.shape)
 
+    @property
+    def live_indices(self):
+        '''Surviving indices that are not eliminated during tensor
+        contractions.'''
+        return self._static_analyzed.live_indices
+
     @property
     def einspec(self):
+        '''NumPy-style einsum specification of the top level operation.'''
         return self._static_analyzed.einspec
 
 
-class ArithmeticMixin:
+def _as_node(x):
+    if isinstance(x, _ASNode):
+        return x
+    elif isinstance(x, _AST):
+        return x.root
+    elif isinstance(x, Number):
+        return P.literal(value=LiteralValue(x))
+    else:
+        raise TypeError(
+            f'Cannot use {x} of type {type(x)} in a tensor expression.'
+        )
 
-    def __add__(self, rhs):
-        return EinopEx(P.ein(
-            self.root, _BaseEx(rhs).root, 6, 'sum', 'add', None
-        ))
 
-    def __radd__(self, lhs):
-        return EinopEx(P.ein(
-            _BaseEx(lhs).root, self.root, 6, 'sum', 'add', None
-        ))
+def as_tsrex(f):
+    def wrapper(*args, **kwargs):
+        return TsrEx(f(*args, **kwargs))
+    return wrapper
 
-    def __sub__(self, rhs):
-        return EinopEx(P.ein(
-            self.root, _BaseEx(rhs).root, 6, 'sum', 'subtract', None
-        ))
 
-    def __rsub__(self, lhs):
-        return EinopEx(P.ein(
-            _BaseEx(lhs).root, self.root, 6, 'sum', 'subtract', None
-        ))
+def yield_tsrex(f):
+    def wrapper(*args, **kwargs):
+        for n in f(*args, **kwargs):
+            yield TsrEx(n)
+    return wrapper
 
-    def __mul__(self, rhs):
-        return EinopEx(P.ein(
-            self.root, _BaseEx(rhs).root, 5, 'sum', 'multiply', None
-        ))
 
-    def __rmul__(self, lhs):
-        return EinopEx(P.ein(
-            _BaseEx(lhs).root, self.root, 5, 'sum', 'multiply', None
-        ))
-
-    def __truediv__(self, rhs):
-        return EinopEx(P.ein(
-            self.root, _BaseEx(rhs).root, 5, 'sum', 'divide', None
-        ))
-
-    def __rtruediv__(self, lhs):
-        return EinopEx(P.ein(
-            _BaseEx(lhs).root, self.root, 5, 'sum', 'divide', None
-        ))
+class SyntaxOverloadMixin:
 
+    @as_tsrex
     def __neg__(self):
-        return TsrEx(P.neg(self.root))
+        return _neg(_as_node(self))
 
-    def __pow__(self, exponent):
-        return TsrEx(P.pow(self.root, _BaseEx(exponent).root))
+    @as_tsrex
+    def __add__(self, rhs):
+        return _binary(_as_node(self), _as_node(rhs), 6, 'add')
 
-    def __rpow__(self, base):
-        return TsrEx(P.pow(_BaseEx(base).root, self.root))
+    @as_tsrex
+    def __sub__(self, rhs):
+        return _binary(_as_node(self), _as_node(rhs), 6, 'subtract')
 
+    @as_tsrex
+    def __mul__(self, rhs):
+        return _binary(_as_node(self), _as_node(rhs), 5, 'multiply')
 
-class IndexRenamingMixin:
-    '''Rename the free indices of a tensor expression.'''
+    @as_tsrex
+    def __matmul__(self, rhs):
+        return _binary(_as_node(self), _as_node(rhs), 5, 'matmul')
 
+    @as_tsrex
+    def __truediv__(self, rhs):
+        return _binary(_as_node(self), _as_node(rhs), 5, 'divide')
+
+    @as_tsrex
+    def __pow__(self, rhs):
+        return _binary(_as_node(self), _as_node(rhs), 3, 'float_power')
+
+    @as_tsrex
+    def __and__(self, rhs):
+        return _binary(_as_node(self), _as_node(rhs), 5, 'kron')
+
+    @as_tsrex
+    def __radd__(self, lhs):
+        return _binary(_as_node(lhs), _as_node(self), 6, 'add')
+
+    @as_tsrex
+    def __rsub__(self, lhs):
+        return _binary(_as_node(lhs), _as_node(self), 6, 'subtract')
+
+    @as_tsrex
+    def __rmul__(self, lhs):
+        return _binary(_as_node(lhs), _as_node(self), 5, 'multiply')
+
+    @as_tsrex
+    def __rmatmul__(self, lhs):
+        return _binary(_as_node(lhs), _as_node(self), 5, 'matmul')
+
+    @as_tsrex
+    def __rtruediv__(self, lhs):
+        return _binary(_as_node(lhs), _as_node(self), 5, 'divide')
+
+    @as_tsrex
+    def __rpow__(self, lhs):
+        return _binary(_as_node(lhs), _as_node(self), 3, 'float_power')
+
+    @as_tsrex
+    def __rand__(self, lhs):
+        return _binary(_as_node(lhs), _as_node(self), 5, 'kron')
+
+    @as_tsrex
     def __getitem__(self, indices):
+        return _getitem(_as_node(self), indices)
 
-        tsrex = self | IndexPropagator()
-
-        indices = as_tuple(indices)
-        live_old = tsrex.root.live_indices
-        if len(indices) != len(live_old):
-            raise SyntaxError(
-                f'Incorrect number of indices. '
-                f'Expects {len(live_old)}, '
-                f'got {len(indices)}.'
-            )
-
-        for new_expr in indices:
-            if new_expr.root.name != 'index':
-                raise SyntaxError(
-                    'Indices to a tensor expression must be abstract indices.'
-                )
-        live_new = [i.root.item for i in indices]
-
-        index_map = dict(zip(live_old, live_new))
-        # if a 'new' live index is already used as a dummy one, replace the
-        # dummy usage with an anonymous index to avoid conflict.
-        for n in dfs_filter(lambda n: n.name == 'index', tsrex.root):
-            i = n.item
-            if i not in live_old and i in live_new:
-                index_map[i] = AbstractIndex()
-
-        for n in dfs_filter(lambda n: n.name == 'index', tsrex.root):
-            n.item = index_map.get(n.item, n.item)
-
-        return tsrex | IndexPropagator()
-
-
-class TranspositionMixin:
-    '''transpose the axes by permuting the live indices into target indices.'''
+    @as_tsrex
     def __rshift__(self, indices):
-        return TsrEx(P.tran(
-            self.root,
-            P.indices(tuple([i.root for i in as_tuple(indices)]))
-        ))
+        return _rshift(_as_node(self), indices)
 
-
-class TsrEx(_BaseEx, ArithmeticMixin, IndexRenamingMixin, TranspositionMixin):
-    '''A general tensor expression'''
-    pass
-
-
-class IndexEx(_BaseEx):
+    @as_tsrex
     def __invert__(self):
-        '''Implements the `~i` syntax.'''
-        return IndexEx(dataclasses.replace(self.root, bound=True, kron=False))
+        return _invert(_as_node(self))
 
+    @yield_tsrex
     def __iter__(self):
-        '''Implements the `*i` syntax.'''
-        yield IndexEx(dataclasses.replace(self.root, bound=False, kron=True))
+        return _iter(_as_node(self))
 
 
-class TensorEx(_BaseEx):
-    def __getitem__(self, indices):
-        return TsrEx(
-            P.index_notation(
-                self.root,
-                P.indices(
-                    tuple([i.root for i in as_tuple(indices or [])])
-                )
-            )
-        )
+class TsrEx(
+    _BaseEx,
+    SyntaxOverloadMixin,
+):
+    '''A expression of potentially nested and composite tensor computations.
+
+    Supported operations between tensor expressions include:
+
+    - Elementwise: addition, subtraction, multiplication, division,
+    exponentiation
+    - Einstein summation and generalized semiring operations
+    - Kronecker product
+    - Transposition
+    - Function evaluation
+
+    For more details, see the corresponding
+    [user guide page](../../pages/user-guide/tsrex).
+    '''
 
 
-class EinopEx(TsrEx):
-    # override the `>>` behavior from TranspositionMixin
-    def __rshift__(self, output_indices):
-        tsrex = EinopEx(copy(self.root))
-        tsrex.root.outidx = P.indices(
-            tuple([i.root for i in as_tuple(output_indices)])
-        )
-        return tsrex
+_dispatch = Dispatcher()
+
+
+@_dispatch
+def _binary(lhs: _ASNode, rhs: _ASNode, precedence, operator):
+    return P.abstract_binary(lhs, rhs, precedence, operator)
+
+
+@_dispatch
+def _neg(node: _ASNode):
+    return P.neg(node)
+
+
+@_dispatch
+def _invert(node: P.index):
+    '''Implements the `~i` syntax.'''
+    return dataclasses.replace(node, bound=True, kron=False)
+
+
+@_dispatch
+def _iter(node: P.index):
+    '''Implements the `*i` syntax.'''
+    yield dataclasses.replace(node, bound=False, kron=True)
+
+
+@_dispatch
+def _getitem(node: _ASNode, indices):  # noqa: F811
+    '''create index notation'''
+    return P.abstract_index_notation(
+        node,
+        P.indices(tuple([i.root for i in as_tuple(indices or [])]))
+    )
+
+
+@_dispatch
+def _rshift(node: _ASNode, indices):  # noqa: F811
+    '''transpose or einsum output specification'''
+    return P.abstract_dest(
+        node,
+        P.indices(tuple([i.root for i in as_tuple(indices)]))
+    )
 
 
 def index(symbol=None):
@@ -272,7 +362,7 @@ def index(symbol=None):
     Returns:
         TsrEx: A single-index tensor expression.
     '''
-    return IndexEx(P.index(AbstractIndex(symbol), bound=False, kron=False))
+    return TsrEx(P.index(AbstractIndex(symbol), bound=False, kron=False))
 
 
 def indices(spec):
@@ -313,7 +403,7 @@ def indices(spec):
         raise RuntimeError(f'Cannot create indices from {spec}.')
 
 
-def tensor(*spec, initializer=None, optimizable=None):
+def tensor(*spec, initializer=None, optimizable=None, prefer=None):
     '''Construct an abstract tensor using `spec`.
 
     Args:
@@ -341,6 +431,11 @@ def tensor(*spec, initializer=None, optimizable=None):
 
             The default behavior can be overriden by user input.
 
+        prefer (callable):
+            Condition evaluated on tensor as a penalty term. Only considered if
+            optimizable is set to True or defaults to True, otherwise it is
+            ignored.
+
     Returns:
         TsrEx: A tensor expression representing an abstract tensor object.
     '''
@@ -375,8 +470,14 @@ def tensor(*spec, initializer=None, optimizable=None):
             raise RuntimeError(
                 f'Tensor size must be positive integer, got {d} instead.'
             )
+    if optimizable and prefer is None:
+        prefer = NoCondition()
 
-    return TensorEx(P.tensor(
-        AbstractTensor(*size, symbol=symbol, initializer=initializer,
-                       optimizable=optimizable))
-                    )
+    return TsrEx(
+        P.tensor(
+            AbstractTensor(
+                *size, symbol=symbol, initializer=initializer,
+                optimizable=optimizable, prefer=prefer
+            )
+        )
+    )
