@@ -4,5 +4,5 @@ from funfact.lang.interpreter._evaluation import Evaluator
 
 
 class ElementwiseEvaluator(Evaluator):
-    def tensor(self, abstract, data, slices, **kwargs):
+    def tensor(self, decl, data, slices, **kwargs):
         return data[tuple(slices)]
