@@ -1,13 +1,13 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 import pytest
-from ._ast import _ASNode, _AST, Primitives
+from ._ast import _ASNode, primitive, _AST
 
 
-def test_primitive_decator():
+def test_make_primitive():
 
     # create new primitive
-    @Primitives.primitive(precedence=1)
+    @primitive(precedence=1)
     def prim(arg1, arg2):
         pass
 
@@ -22,6 +22,7 @@ def test_primitive_decator():
 
     # check for node instance properties
     node = prim('a', False)
+    assert isinstance(node, prim)
     assert node.name == 'prim'
     assert node.precedence == 1
     assert 'arg1' in node.fields_fixed
@@ -36,31 +37,22 @@ def test_primitive_decator():
     assert node.payload1 == 1
     assert node.payload2 == 1.5
 
+
+def test_make_primitive_dynamic_precedence():
     # per-node precedence
-    @Primitives.primitive()
-    def prim_nopred(precedence, arg1, arg2):
+    @primitive()
+    def prim_dynamic_pred(precedence, arg1, arg2):
         pass
 
-    assert not hasattr(prim_nopred, 'precedence')
+    assert not hasattr(prim_dynamic_pred, 'precedence')
 
-    node = prim_nopred(2, 'a', True)
-    assert node.name == 'prim_nopred'
+    node = prim_dynamic_pred(2, 'a', True)
+    assert node.name == 'prim_dynamic_pred'
     assert node.precedence == 2
 
 
 def test_ast():
-
-    @Primitives.primitive()
-    def prim():
-        pass
-
-    # constract from scalar
-    ast = _AST(1)
+    ast = _AST(None)
     assert hasattr(ast, 'root')
-    assert ast.root.name == 'scalar'
-    # re-assign root
-    ast.root = prim()
-    assert ast.root.name == 'prim'
-
-    # construct from ASNode
-    assert _AST(prim()).root.name == 'prim'
+    assert ast.root is None
+    ast.root = None  # check if root is writable
