@@ -2,14 +2,15 @@
 # -*- coding: utf-8 -*-
 from funfact.lang.interpreter import (
     dfs_filter,
+    Compiler,
     EinsteinSpecGenerator,
     Evaluator,
+    IndexnessAnalyzer,
     LeafInitializer,
-    IndexPropagator,
     ElementwiseEvaluator,
     SlicingPropagator,
-    ShapeAnalyzer
 )
+from funfact import active_backend as ab
 
 
 class Factorization:
@@ -36,22 +37,23 @@ class Factorization:
 
     def __init__(self, tsrex, **extra_attributes):
         self._tsrex = (tsrex
-                       | IndexPropagator()
-                       | EinsteinSpecGenerator()
-                       | ShapeAnalyzer())
+                       | IndexnessAnalyzer()
+                       | Compiler()
+                       | EinsteinSpecGenerator())
         self.__dict__.update(**extra_attributes)
 
     @classmethod
-    def from_tsrex(cls, tsrex, initialize=True):
+    def from_tsrex(cls, tsrex, dtype=None, initialize=True):
         '''Construct a factorization model from a tensor expresson.
 
         Args:
             tsrex (TsrEx): The tensor expression.
+            dtype: numerical data type, defaults to float32.
             initialize (bool):
                 Whether or not to fill abstract tensors with actual data.
         '''
         if initialize:
-            tsrex = tsrex | LeafInitializer()
+            tsrex = tsrex | LeafInitializer(dtype)
         return cls(tsrex)
 
     @property
@@ -78,14 +80,14 @@ class Factorization:
         return self._NodeView(
             'data',
             list(dfs_filter(lambda n: n.name == 'tensor' and
-                            n.abstract.optimizable, self.tsrex.root))
+                            n.decl.optimizable, self.tsrex.root))
         )
 
     @factors.setter
     def factors(self, tensors):
         for i, n in enumerate(
             dfs_filter(lambda n: n.name == 'tensor' and
-                       n.abstract.optimizable, self.tsrex.root)
+                       n.decl.optimizable, self.tsrex.root)
         ):
             n.data = tensors[i]
 
@@ -132,6 +134,25 @@ class Factorization:
         '''The dimensionality of the result tensor.'''
         return self.tsrex.ndim
 
+    def penalty(self, sum_leafs: bool = True, sum_vec=None):
+        '''The penalty of the result tensor.
+
+        Args:
+            sum_leafs (bool): sum the penalties over the leafs of the model.
+            sum_vec (bool): sum the penalties over the vectorization dimension.
+        '''
+
+        factors = list(dfs_filter(
+                lambda n: n.name == 'tensor' and n.decl.optimizable,
+                self.tsrex.root)
+        )
+        penalties = ab.stack(
+            [f.decl.prefer(f.data, sum_vec) for f in factors],
+            0 if sum_vec else -1
+        )
+        return ab.sum(penalties, 0 if sum_vec else -1) if sum_leafs else \
+            penalties
+
     def __call__(self):
         '''Shorthand for :py:meth:`forward`.'''
         return self.forward()
@@ -193,7 +214,7 @@ class Factorization:
         elements.'''
         if isinstance(idx, str):
             for n in dfs_filter(
-                lambda n: n.name == 'tensor' and str(n.abstract.symbol) == idx,
+                lambda n: n.name == 'tensor' and str(n.decl.symbol) == idx,
                 self.tsrex.root
             ):
                 return n.data
@@ -204,7 +225,7 @@ class Factorization:
     def __setitem__(self, name, data):
         '''Implements attribute-based access of factor tensors.'''
         for n in dfs_filter(
-            lambda n: n.name == 'tensor' and str(n.abstract.symbol) == name,
+            lambda n: n.name == 'tensor' and str(n.decl.symbol) == name,
             self.tsrex.root
         ):
             return setattr(n, 'data', data)
@@ -218,7 +239,7 @@ class Factorization:
         def __repr__(self):
             return '<{attr} field{pl} of tensor{pl} {tensors}>'.format(
                 attr=repr(self.attribute),
-                tensors=', '.join([str(n.abstract) for n in self.nodes]),
+                tensors=', '.join([str(n.decl) for n in self.nodes]),
                 pl='s' if len(self.nodes) > 1 else ''
             )
 
